/* 
 * gimp_plugin.c is part of the Verse Gimp Plugin.
 * Copyright (C) 2003 Brecht Van Lommel <blendix@pandora.be>
 *
 * Verse Gimp Plugin is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Verse Gimp Plugin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Verse Gimp Plugin; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <verse.h>
#undef DBL_MAX
#undef TRUE
#undef FALSE
#include <stdlib.h>
#include <stdio.h>
#include <gtk/gtk.h>
#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

#include "verse_gimp.h"

static void query();
static void run(const gchar *name, gint nparams, const GimpParam *param, gint *nreturn_vals, GimpParam **return_vals);

GimpPlugInInfo PLUG_IN_INFO =
{
	NULL,  /* init_proc  */
	NULL,  /* quit_proc  */
	query, /* query_proc */
	run,   /* run_proc   */
};

/* Give the Gimp the info it asks for. */
static void query()
{
	static GimpParamDef args[] =
	{
		{ GIMP_PDB_INT32, "run_mode", "Interactive, [non-interactive]" }
	};
	
	gimp_install_procedure("verse_gimp",
	                       "Connect Gimp with a Verse Network.",
                           "",
                           "Brecht Van Lommel <blendix@pandora.be>",
                           "Brecht Van Lommel <blendix@pandora.be>",
                           "2003",
                           "<Toolbox>/Xtns/_Verse Interaction",
                           "RGB*, GRAY*, INDEXED*",
                           GIMP_PLUGIN,
                           G_N_ELEMENTS(args), 0,
                           args, NULL);
}

/* Initialize the plugin. */
static void run (const gchar *name,
                 gint nparams,
                 const GimpParam  *param,
                 gint *nreturn_vals,
                 GimpParam **return_vals)
{
	static GimpParam values[1];
	GimpRunMode      run_mode;
	run_mode = param[0].data.d_int32;
	*nreturn_vals = 1;
	*return_vals  = values;
	values[0].type          = GIMP_PDB_STATUS;
	values[0].data.d_status = GIMP_PDB_SUCCESS;

	if(run_mode == GIMP_RUN_INTERACTIVE)
		gui_init();
}

/* This macro replaces the normal main function. */
MAIN()

