/* 
 * verse_connection.c is part of the Verse Gimp Plugin.
 * Copyright (C) 2003 Brecht Van Lommel <blendix@pandora.be>
 *
 * Verse Gimp Plugin is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Verse Gimp Plugin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Verse Gimp Plugin; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#undef DBL_MAX
#undef TRUE
#undef FALSE
#include <verse.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <libgimp/gimpmisc_pdb.h>

#include "verse_connection.h"
#include "verse_gimp.h"

int vrs_callback_update()
{
	verse_callback_update(1000/*gui_get_verse_update_interval()*/);
	return TRUE;
}

void cb_connect_accept(void *user, VNodeID avatar, const char *address, uint8 *hostid)
{
	/* store avatar's id vg data struct */
	gui_set_avatar(avatar);
	/* set up connection state */
	gui_set_state_connected();
	/* verse gimp will receive changes of bitmap and object nodes, because
	 * one object node represents gimp as verse avatar */
	verse_send_node_index_subscribe(1 << V_NT_BITMAP | 1 << V_NT_OBJECT);
}

void cb_node_create(void *user, VNodeID node_id, uint8 type, VNodeOwner owner)
{
	i_bitmap_t *bitmap;
	gint32 id;

	/* send name of verse avatar to verse server */
	if (type == V_NT_OBJECT && node_id == gui_get_avatar()) {
		gchar *base_name = "gimp_";
		gchar *name = NULL;
		gchar *version = gimp_version();
		if(version) {
			printf("base_name: %s\n", base_name);
			printf("version: %s\n", version);
			name = malloc(sizeof(gchar)*(strlen(base_name)+strlen(version)));
			name[0]='\0';
			strcat(name, base_name);
			strcat(name, version);
		}
		verse_send_node_name_set(node_id, name);
		if(name) free(name);
	}
	
	if (type != V_NT_BITMAP)
		return;

	if ((bitmap = add_bitmap_from_verse(node_id)) == NULL)
		return;

	verse_send_node_subscribe(node_id);

	if ((owner == VN_OWNER_MINE) && (id = pop_id_upload_queue())) {
		if (set_bitmap_from_gimp(bitmap, id)) {
			verse_send_b_dimensions_set(node_id, bitmap->width, bitmap->height, 1);
			verse_send_node_name_set(node_id, bitmap->name);
		} else
			verse_send_node_destroy(node_id);
	}

	gui_refresh_verse_list();
}

void cb_node_destroy(void *user, VNodeID node_id)
{
	i_bitmap_t *bitmap;
	if ((bitmap = get_bitmap(node_id)) != NULL)
	{
		remove_bitmap(bitmap);
		gui_refresh_verse_list();
	}
}

void cb_node_name_set(void *user, VNodeID node_id, char *name)
{
	i_bitmap_t *bitmap;
	if ((bitmap = get_bitmap(node_id)) != NULL)
	{
		if (bitmap->name != NULL)
			g_free(bitmap->name);
		bitmap->name = g_strdup(name);
		gui_refresh_verse_list();
	}
}

static void cb_dimensions(void *user_data, VNodeID node_id, uint16 width, uint16 height, uint16 depth)
{
	i_bitmap_t *bitmap;
	if ((bitmap = get_bitmap(node_id)) != NULL) {
		/* FIXME this is a hack, but we have to do it, because verse sends us
		   0x0, after the real size. Probably a verse bug. Or not? */
		if (width == 0 || height == 0)
			return;

		/* FIXME changing bitmap size might be fatal */
		if (bitmap->width != width || bitmap->height != height)
			resize_bitmap(bitmap, width, height);

		if (bitmap->creation_pending) {
			verse_send_b_layer_create(node_id, -1, "col_r", NATIVE_LAYER_TYPE);
			verse_send_b_layer_create(node_id, -2, "col_g", NATIVE_LAYER_TYPE);
			verse_send_b_layer_create(node_id, -3, "col_b", NATIVE_LAYER_TYPE);
			verse_send_b_layer_create(node_id, -4, "alpha", NATIVE_LAYER_TYPE);
			bitmap->creation_pending = FALSE;
		}
		gui_refresh_verse_list();
	}
}

static void cb_layer_create(void *user_data, VNodeID node_id, VLayerID layer_id, char *name, VNBLayerType type)
{
	i_bitmap_t *bitmap;
	i_layer_t *layer;

	if ((bitmap = get_bitmap(node_id)) == NULL ||
		/*bitmap->width == 0 || bitmap->height == 0 ||*/
	    (layer = add_layer_from_verse(bitmap, layer_id, name, type)) == NULL ||
	    !(layer->gimp_channel >= 0))
		return;

	if (bitmap->download)
		verse_send_b_layer_subscribe(node_id, layer_id, 0);
}

static void cb_layer_destroy(void *user_data, VNodeID node_id, VLayerID layer_id)
{
	i_bitmap_t *bitmap;
	i_layer_t *layer;
	if (((bitmap = get_bitmap(node_id)) != NULL) &&
	    ((layer = get_layer(bitmap, layer_id)) != NULL)){
		remove_layer(layer);
	}
}

static void cb_tile(void *user_data, VNodeID node_id, VLayerID layer_id, uint16 tile_x, uint16 tile_y, uint16 z, VNBLayerType type, VNBTile *data)
{
	i_bitmap_t *bitmap;
	i_layer_t *layer;
	static int got = 0;
	got++;
	/*printf("got: %d\n", got);*/

	if (((bitmap = get_bitmap(node_id)) != NULL) && bitmap->download && bitmap->width > 0 && bitmap->height > 0)
	{
		if ((layer = get_layer(bitmap, layer_id))!=NULL)
		{
			if(layer->data != NULL) 
				put_layer_data(layer, tile_x, tile_y, data, type);
		}
	}
#if 0
			else
				exit(0);
		}
		else
			exit(0);
	}
	else
		exit(0);
#endif
}

VSession *vrs_connect(char *address, char *uname, char *pass)
{
	VSession *session;

	verse_callback_set(verse_send_connect_accept, cb_connect_accept, NULL);
	verse_callback_set(verse_send_node_create, cb_node_create, NULL);
	verse_callback_set(verse_send_node_destroy, cb_node_destroy, NULL);
	verse_callback_set(verse_send_node_name_set, cb_node_name_set, NULL);
	verse_callback_set(verse_send_b_dimensions_set, cb_dimensions, NULL);
	verse_callback_set(verse_send_b_layer_create, cb_layer_create, NULL);
	verse_callback_set(verse_send_b_layer_destroy, cb_layer_destroy, NULL);
	verse_callback_set(verse_send_b_tile_set, cb_tile, NULL);
	session = verse_send_connect(uname, pass, address, NULL);
	verse_session_set(session);
	gui_set_state_connecting();
	return session;
}

void vrs_disconnect(VSession *session, char *address, char *message)
{
	verse_send_connect_terminate(address, message);
	verse_session_destroy(session);
	gtk_timeout_remove(gui_get_verse_timeout());
	gui_destroy_server_data();
	gui_set_state_disconnecting();
}

void vrs_create_bitmap()
{
	static int i = 0;
	verse_send_node_create(i--, V_NT_BITMAP, gui_get_avatar());
}

