/* 
 * verse_gimp.h is part of the Verse Gimp Plugin.
 * Copyright (C) 2003 Brecht Van Lommel <blendix@pandora.be>
 *
 * Verse Gimp Plugin is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Verse Gimp Plugin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Verse Gimp Plugin; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#define TILES(length) ((length + VN_B_TILE_SIZE - 1) / VN_B_TILE_SIZE)
/* XXX assuming gimp_tile_width = gimp_tile_height */
#define NTILES(length) ((length + gimp_tile_width() - 1) / gimp_tile_width())
#define NTILE_SIZE gimp_tile_width()
#define ONE_TILE (VN_B_TILE_SIZE * VN_B_TILE_SIZE)
#define NATIVE_LAYER_TYPE VN_B_LAYER_UINT8

#define MAX_TILES 3000
#define MAX_CALLBACK_UPDATE 100
#define INTERVAL_CALLBACK_UPDATE 200

typedef struct {
	gboolean upload;
	gboolean download;
	gboolean creation_pending;
	gboolean redraw_pending;
	guint current_tile;
	GArray *redraw_queue;

	gint32 gimp_image_id;
	gint32 gimp_layer_id;
	VNodeID verse_node_id;

	uint16 width;
	uint16 height;
	gchar *name;
	GList *layers;
} i_bitmap_t;

typedef struct {
	uint8 *data;
	guint w;
	guint h;
	gboolean dirty;
} i_tile_t;

typedef struct {
	i_bitmap_t *parent;
	guint32 verse_layer_id;

	VNBLayerType verse_layer_type;
	gint gimp_channel; /* -1=None, 0=R, 1=G, 2=B, 3=Alpha */
	char *name;
	i_tile_t *data; 
} i_layer_t;

void gui_set_avatar(guint32 avatar);
guint32 gui_get_avatar();
guint gui_get_verse_timeout();
guint gui_get_verse_update_interval();

uint8 *get_layer_data_from_gimp(i_layer_t *layer, uint16 tile_x, uint16 tile_y);
void put_layer_data(i_layer_t *layer, uint16 tile_x, uint16 tile_y, VNBTile *data, VNBLayerType type);
void draw_layer_data(i_layer_t *layer, uint16 tile_x, uint16 tile_y, uint8 *data);
void update_layer_data(i_layer_t *layer);

gint32 pop_id_upload_queue();
guint pop_redraw_queue(i_bitmap_t *bitmap);
i_bitmap_t *get_bitmap(VNodeID id);
i_bitmap_t *get_bitmap_verse_id(gint32 id);
i_layer_t *get_layer(i_bitmap_t *bitmap, VLayerID layer_id);
i_bitmap_t *add_bitmap_from_verse(VNodeID id);
i_layer_t *add_layer_from_verse(i_bitmap_t *bitmap, VLayerID layer_id, char *name, VNBLayerType type);
gboolean set_bitmap_from_gimp(i_bitmap_t *bitmap, gint32 gimp_image_id);
void remove_bitmap(i_bitmap_t *bitmap);
void unlink_bitmap(i_bitmap_t *bitmap);
void remove_layer(i_layer_t *layer);
void resize_bitmap(i_bitmap_t *bitmap, uint new_width, uint new_height);

void gui_refresh_verse_list();
void gui_destroy_server_data();
void gui_init();
void gui_set_state_connecting();
void gui_set_state_connected();
void gui_set_state_disconnected();
void gui_set_state_disconnecting();

