
#include "verse_ms.h"

#include "verse_gimp.h"

struct Global G;

/* Refresh */

void gui_refresh(void)
{
	/* delay untill after all updates are done */
	G.gui.need_refresh = TRUE;
}

void gui_do_refresh(void)
{
	GList *i;
	VGBitmap *bmp;
	GtkTreeIter iter;
	char *name, *gimp_id;
	guint32 image_id;
	VNodeID node_id;

	gtk_list_store_clear(G.gui.gimplist);
	gtk_list_store_clear(G.gui.verselist);

	for (i = G.sync.bitmaps; i; i = g_list_next(i)) {
		bmp = (VGBitmap*)i->data;
		image_id = (bmp->g.sync)? bmp->g.image_id: 0;

		if (bmp->g.sync) {
			name = bmp->name;

			gtk_list_store_append(G.gui.gimplist, &iter);
			gtk_list_store_set(G.gui.gimplist, &iter, 0, image_id, 1, name, -1);
		}
		if (bmp->v.sync) {
			gtk_list_store_append(G.gui.verselist, &iter);

			node_id = bmp->v.node_id;

			if (bmp->name)
				name = g_strdup_printf("%s (%ix%i)", bmp->name, bmp->width, bmp->height);
			else
				name = g_strdup_printf("*** (%ix%i)", bmp->width, bmp->height);

			if (bmp->g.sync)
				gimp_id = g_strdup_printf("%d", image_id);
			else
				gimp_id = g_strdup_printf(" ");

			gtk_list_store_set(G.gui.verselist, &iter, 0, node_id, 1, name,
			                   2, gimp_id, -1);
			g_free(name);
			g_free(gimp_id);
		}
	}
}



/* Callbacks */

static gboolean gui_stop(gpointer data)
{
	gui_set_conn_state(DISCONNECTED);
	gtk_main_quit();

	return FALSE;
}

static void cb_gui_quit(GtkObject *object, gpointer user_data)
{	
	if ((G.gui.state == CONNECTED) || (G.gui.state == CONNECTING))
		gui_set_conn_state(DISCONNECTED);
	
	if (G.gui.state != DISCONNECTED)
		g_timeout_add(500, gui_stop, NULL);
	else
		gui_stop(NULL);
}

static void cb_gui_connect(GtkObject *object, gpointer user_data)
{	
	gui_set_conn_state(CONNECTING);
}

static void cb_gui_disconnect(GtkObject *object, gpointer user_data)
{	
	gui_set_conn_state(DISCONNECTED);
}

static gboolean cb_gui_update(gpointer data)
{
	/* always do this first to make sure all VGBitmap's are still valid */
	if (G.gui.state == CONNECTED)
		sync_gimp_bitmaps();
	/* then do verse callback update */
	conn_callback_update();
	if (G.gui.state == CONNECTED) {
		static int counter = 0;

		sync_gimp_bitmaps();


		if (counter == 0) {
			/* update gimp pixels */
			sync_pixels(100);
		}

		counter = (counter + 1)%50;

		if (G.gui.need_refresh) {
			G.gui.need_refresh = FALSE;
			gui_do_refresh();
		}
	}
	return TRUE; /* call again */
}

static void cb_gui_ms_refresh(GtkWidget *widget, gpointer data)
{
	const char	*ip;

	if((ip = gtk_entry_get_text(GTK_ENTRY(G.gui.ms_address))) == NULL)
		return;
	gtk_list_store_clear(GTK_LIST_STORE(G.gui.ms_store));
	verse_ms_get_send(ip, VERSE_MS_FIELD_DESCRIPTION, NULL);
}

static void cb_gui_ms_activated(GtkWidget *widget, GtkTreePath *path, GtkTreeViewColumn *col, gpointer user)
{
	GtkTreeIter   iter;

	if(gtk_tree_model_get_iter(GTK_TREE_MODEL(G.gui.ms_store), &iter, path))
	{
		gchar *name;

		gtk_tree_model_get(GTK_TREE_MODEL(G.gui.ms_store), &iter, 0, &name, -1);
		gtk_entry_set_text(GTK_ENTRY(G.gui.address), name);
		g_free(name);
	}
}

static VGBitmap *gui_get_selected_bitmap(GtkWidget *tree)
{
	GtkTreeIter iter;
	GtkTreeModel *model;
	GtkTreeSelection *select;
	gint32 id;

	select = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));
	if (gtk_tree_selection_get_selected (select, &model, &iter)) {
		gtk_tree_model_get(model, &iter, 0, &id, -1);

		if (tree == G.gui.versetree)
			return sync_bitmap_lookup_verse(id);
		else
			return sync_bitmap_lookup_gimp(id);
	}

	return NULL;
}

static void cb_gui_sync(GtkWidget *widget, gpointer data)
{
	VGBitmap *bmp = gui_get_selected_bitmap(G.gui.gimptree);

	if (bmp)
		sync_gimp_bitmap_sync(bmp);
}

static void cb_gui_destroy(GtkWidget *widget, gpointer data)
{
	VGBitmap *bmp = gui_get_selected_bitmap(G.gui.versetree);

	if (bmp)
		sync_gimp_bitmap_destroy(bmp);
}

static void cb_gui_subscribe(GtkWidget *widget, gpointer data)
{
	VGBitmap *bmp = gui_get_selected_bitmap(G.gui.versetree);

	if (bmp)
		sync_gimp_bitmap_subscribe(bmp);
}

static void cb_gui_unsubscribe(GtkWidget *widget, gpointer data)
{
	VGBitmap *bmp = gui_get_selected_bitmap(G.gui.versetree);

	if (bmp)
		sync_gimp_bitmap_unsubscribe(bmp);
}

static void cb_ping(void *user, const char *address, const char *msg)
{
	VMSServer	**s;

	if(G.gui.ms_store == NULL)
		return;
	if((s = verse_ms_list_parse(msg)) != NULL)
	{
		int		i;
		GtkTreeIter	iter;
		const char	*de;

		for(i = 0; s[i] != NULL; i++)
		{
			de = verse_ms_field_value(s[i], VERSE_MS_FIELD_DESCRIPTION_NAME);
			if(de == NULL)
				de = "";
			gtk_list_store_append(G.gui.ms_store, &iter);
			gtk_list_store_set(G.gui.ms_store, &iter,
					   0, s[i]->ip,
					   1, de,
					   -1);

		}
		free(s);
	}
}

/* Setting states */

void gui_set_conn_state(enum ConnectionState state)
{
	GtkWidget *page = gtk_notebook_get_nth_page(GTK_NOTEBOOK(G.gui.notebook), 1);

	if (G.gui.state == state)
		return;

	if (state == CONNECTING) {
		const char *addr = (char*)gtk_entry_get_text(GTK_ENTRY(G.gui.address));
		const char *uname = (char*)gtk_entry_get_text(GTK_ENTRY(G.gui.uname));
		const char *pass = (char*)gtk_entry_get_text(GTK_ENTRY(G.gui.pass));

		gtk_widget_set_sensitive(G.gui.table_server, FALSE);
		gtk_widget_set_sensitive(G.gui.conn, FALSE);
		gtk_widget_set_sensitive(G.gui.disconn, TRUE);
		gtk_widget_set_sensitive(page, FALSE);
		gtk_label_set_markup(GTK_LABEL(G.gui.state_label), "State: <b>connecting</b>");

		/* force gtk gui update, verse connect can take a long time */
		gtk_main_iteration();

		conn_start(addr, uname, pass);
	}
	else if (state == CONNECTED) {
		gtk_widget_set_sensitive(G.gui.table_server, FALSE);
		gtk_widget_set_sensitive(G.gui.conn, FALSE);
		gtk_widget_set_sensitive(G.gui.disconn, TRUE);
		gtk_widget_set_sensitive(page, TRUE);
		gtk_label_set_markup(GTK_LABEL(G.gui.state_label), "state: <b>connected</b>");

		sync_start();
		gui_refresh();
	}
	else if (state == DISCONNECTED) {
		gtk_widget_set_sensitive(G.gui.table_server, TRUE);
		gtk_widget_set_sensitive(G.gui.conn, TRUE);
		gtk_widget_set_sensitive(G.gui.disconn, FALSE);
		gtk_notebook_set_current_page(GTK_NOTEBOOK(G.gui.notebook), 0);
		gtk_widget_set_sensitive(page, FALSE);
		gtk_label_set_markup(GTK_LABEL(G.gui.state_label), "state: <b>disconnected</b>");

		if (G.gui.state != UNINITIALIZED) {
			conn_stop();
			sync_stop();
			gui_refresh();
		}
	}

	G.gui.state = state;
}

enum ConnectionState gui_get_conn_state(void)
{
	return G.gui.state;
}

/* Create widgets and Run */

static void gui_create_window(void)
{
	GtkWidget *vbox, *vbox2, *hbox, *label, *subscribe;
	GtkWidget *unsubscribe, *sync, *destroy, *topvbox, *scwin;
	GtkWidget *ex, *msb, *mshb, *msbut;
	GtkCellRenderer	*cr;
	GtkTreeViewColumn	*col;

	/* The Dialog and notebook. */
	gimp_ui_init("verse_gimp", FALSE);
	G.gui.window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(G.gui.window), "Verse");
	gtk_window_set_role(GTK_WINDOW(G.gui.window), "gimp-2.2");
	topvbox = gtk_vbox_new(FALSE, 6);
	gtk_container_add(GTK_CONTAINER(G.gui.window), topvbox);
	
	g_signal_connect(G.gui.window, "destroy", G_CALLBACK(cb_gui_quit), NULL);
	G.gui.notebook = gtk_notebook_new();
	gtk_box_pack_start(GTK_BOX(topvbox), G.gui.notebook, TRUE, TRUE, 0);

	/* Connection Settings */
	vbox2 = gtk_vbox_new(FALSE, 3);

	G.gui.table_server = gtk_table_new(4, 2, FALSE);
	gtk_table_set_col_spacings(GTK_TABLE(G.gui.table_server), 4);
	gtk_table_set_row_spacings(GTK_TABLE(G.gui.table_server), 2);
	gtk_container_set_border_width(GTK_CONTAINER(G.gui.table_server), 3);

	/* Master server list. */
	ex = gtk_expander_new ("Master Server");
	msb = gtk_vbox_new(FALSE, 0);
	mshb = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(mshb), gtk_label_new("Master:"), FALSE, FALSE, 0);
	G.gui.ms_address = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(G.gui.ms_address), "master.uni-verse.org:4950");
	gtk_box_pack_start(GTK_BOX(mshb), G.gui.ms_address, TRUE, TRUE, 0);
	msbut = gtk_button_new_from_stock(GTK_STOCK_REFRESH);
	g_signal_connect(G_OBJECT(msbut), "clicked", G_CALLBACK(cb_gui_ms_refresh), NULL);
	gtk_box_pack_start(GTK_BOX(mshb), msbut, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(msb), mshb, FALSE, FALSE, 0);

	G.gui.ms_store = gtk_list_store_new(2, G_TYPE_STRING, G_TYPE_STRING);
	G.gui.ms_list_view = gtk_tree_view_new_with_model(GTK_TREE_MODEL(G.gui.ms_store));
	cr = gtk_cell_renderer_text_new();
	col = gtk_tree_view_column_new_with_attributes("IP", cr, "text", 0, NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(G.gui.ms_list_view), col);
	col = gtk_tree_view_column_new_with_attributes("Description", cr, "text", 1, NULL);
	g_signal_connect(G.gui.ms_list_view, "row-activated", (GCallback) cb_gui_ms_activated, NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(G.gui.ms_list_view), col);
	scwin = gtk_scrolled_window_new(NULL, NULL);
	gtk_container_add(GTK_CONTAINER(scwin), G.gui.ms_list_view);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scwin), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_box_pack_start(GTK_BOX(msb), scwin, TRUE, TRUE, 0);

	gtk_container_add(GTK_CONTAINER(ex), msb);
	gtk_box_pack_start (GTK_BOX(vbox2), ex, TRUE, TRUE, 0);

	/* Server address and login info. */
	label = gtk_label_new("Server:"); 
	gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
	gtk_table_attach(GTK_TABLE(G.gui.table_server), label,
	                         0, 1, 0, 1, 0, 0, 0, 0);
	G.gui.address = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(G.gui.address), "localhost");
	gtk_table_attach_defaults(GTK_TABLE(G.gui.table_server), G.gui.address,
	                         1, 2, 0, 1);

	label = gtk_label_new("Username:"); 
	gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
	gtk_table_attach(GTK_TABLE(G.gui.table_server), label,
	                         0, 1, 2, 3, 0, 0, 0, 0);
	G.gui.uname = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(G.gui.uname), "anonymous");
	gtk_table_attach_defaults(GTK_TABLE(G.gui.table_server), G.gui.uname,
	                         1, 2, 2, 3);

	label = gtk_label_new("Password:"); 
	gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
	gtk_table_attach(GTK_TABLE(G.gui.table_server), label,
	                         0, 1, 3, 4, 0, 0, 0, 0);
	G.gui.pass = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(G.gui.pass), "anonymous");
	gtk_table_attach_defaults(GTK_TABLE(G.gui.table_server), G.gui.pass,
	                         1, 2, 3, 4);

	gtk_box_pack_start(GTK_BOX(vbox2), G.gui.table_server, FALSE, FALSE, 3);
	
	G.gui.state_label = gtk_label_new("");
	gtk_box_pack_start(GTK_BOX(vbox2), G.gui.state_label, FALSE, FALSE, 0);

	label = gtk_label_new("Connection");
	gtk_notebook_append_page(GTK_NOTEBOOK(G.gui.notebook), vbox2, label);

	/* The Gimp Images List tab */
	vbox = gtk_vbox_new(FALSE, 3);

	G.gui.gimplist = gtk_list_store_new(2, G_TYPE_INT, G_TYPE_STRING);
	G.gui.gimptree = gtk_tree_view_new_with_model(GTK_TREE_MODEL(G.gui.gimplist));
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(G.gui.gimptree),
	                                            -1, "Gimp ID",
	                                           gtk_cell_renderer_text_new(),
	                                           "text", 0, NULL);
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(G.gui.gimptree),
	                                            -1, "Name",
	                                            gtk_cell_renderer_text_new(),
	                                            "text", 1, NULL);
	scwin = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scwin),
	                               GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(scwin), G.gui.gimptree);
	gtk_box_pack_start(GTK_BOX(vbox), scwin, TRUE, TRUE, 0);

	hbox = gtk_hbox_new(FALSE, 3);
	sync = gtk_button_new_with_mnemonic("Sync to Verse");
	g_signal_connect(sync, "clicked", G_CALLBACK(cb_gui_sync), NULL);
	gtk_box_pack_start(GTK_BOX(hbox), sync, TRUE, TRUE, 0);

	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

	/* The Verse Bitmaps List tab */
	G.gui.verselist = gtk_list_store_new(3, G_TYPE_INT, G_TYPE_STRING,
	                                     G_TYPE_STRING);
	G.gui.versetree = gtk_tree_view_new_with_model(GTK_TREE_MODEL(G.gui.verselist));
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(G.gui.versetree),
	                                           -1, "Verse ID",
	                                           gtk_cell_renderer_text_new(),
	                                           "text", 0, NULL);
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(G.gui.versetree),
	                                           -1, "Name",
	                                           gtk_cell_renderer_text_new(),
	                                           "text", 1, NULL);
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(G.gui.versetree),
	                                            -1, "Gimp ID",
	                                            gtk_cell_renderer_text_new(),
	                                            "text", 2, NULL);
	scwin = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scwin),
					GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(scwin), G.gui.versetree);
	gtk_box_pack_start(GTK_BOX(vbox), scwin, TRUE, TRUE, 0);

	hbox = gtk_hbox_new(FALSE, 3);
	subscribe = gtk_button_new_with_mnemonic("Subscribe");
	g_signal_connect(subscribe, "clicked", G_CALLBACK(cb_gui_subscribe), NULL);
	gtk_box_pack_start(GTK_BOX(hbox), subscribe, TRUE, TRUE, 0);

	unsubscribe = gtk_button_new_with_mnemonic("Unsubscribe");
	g_signal_connect(unsubscribe, "clicked", G_CALLBACK(cb_gui_unsubscribe), NULL);
	gtk_box_pack_start(GTK_BOX(hbox), unsubscribe, TRUE, TRUE, 0);

	destroy = gtk_button_new_with_mnemonic("Destroy");
	g_signal_connect(destroy, "clicked", G_CALLBACK(cb_gui_destroy), NULL);
	gtk_box_pack_start(GTK_BOX(hbox), destroy, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

	label = gtk_label_new("Images");
	gtk_notebook_append_page(GTK_NOTEBOOK(G.gui.notebook), vbox, label);

	hbox = gtk_hbox_new(FALSE, 3);
	G.gui.conn = gtk_button_new_from_stock(GTK_STOCK_CONNECT);
	g_signal_connect(G.gui.conn, "clicked", G_CALLBACK(cb_gui_connect), NULL);
	gtk_box_pack_start(GTK_BOX(hbox), G.gui.conn, TRUE, TRUE, 0);

	G.gui.disconn = gtk_button_new_from_stock(GTK_STOCK_DISCONNECT);
	g_signal_connect(G.gui.disconn, "clicked", G_CALLBACK(cb_gui_disconnect), NULL);
	gtk_box_pack_start(GTK_BOX(hbox), G.gui.disconn, TRUE, TRUE, 0);

	destroy = gtk_button_new_from_stock(GTK_STOCK_CLOSE);
	g_signal_connect(destroy, "clicked", G_CALLBACK(cb_gui_quit), NULL);
	gtk_box_pack_start(GTK_BOX(hbox), destroy, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(topvbox), hbox, FALSE, FALSE, 0);

	gtk_widget_set_size_request(G.gui.window, 450, 320);
	gtk_widget_show_all(G.gui.window);

	verse_callback_set((void *) verse_send_ping, cb_ping, NULL);
}

static void gui_start(void)
{
	gui_create_window();

	G.gui.state = UNINITIALIZED;
	G.gui.need_refresh = TRUE;
	gui_set_conn_state(DISCONNECTED);

	g_timeout_add(20, cb_gui_update, NULL);
}

void gui_run(void)
{
	gui_start();
	gtk_main();
}

