
#include <string.h>
#include <stdlib.h>
#include "verse_gimp.h"

#include <sys/time.h>

static double timer()
{
	GTimeVal t;
	g_get_current_time(&t);

	return (double)t.tv_sec + ((double)t.tv_usec)/1e6;
}

#define TILE_TIMEOUT 5.0

/* Pixel Operations */

gboolean sync_vtile_pixels_gimp(VGLayer *layer, VTile *vtile, GimpTile *tile, int tx, int ty, int tw, int th, double curtime)
{
	int vtile_rowskip = VN_B_TILE_SIZE - tw;
	int tile_rowskip = (tile->ewidth - tw)*tile->bpp;
	int x, y, channel;
	uint8 *in;
	uint8 *out, *outptr = tile->data + (ty*tile->ewidth + tx)*tile->bpp;
	gboolean skipped = FALSE;

	/* copy one verse tile into a gimp tile */
	for (channel = 0; channel < 4; channel++) {
		if (!vtile->g.need_update[channel])
			continue;
		if (vtile->g.last_diff + TILE_TIMEOUT > curtime) {
			skipped = TRUE;
			continue;
		}
		vtile->g.need_update[channel] = FALSE;

		in = vtile->v.data[channel];
		out = outptr + channel;
		for (y = 0; y < th; y++) {
			for (x = 0; x < tw; x++, in++, out+=tile->bpp)
				*out = *in;

			in += vtile_rowskip;
			out += tile_rowskip;
		}

		memcpy(vtile->g.data[channel], vtile->v.data[channel], sizeof vtile->g.data[channel]);
		vtile->g.uninitialized[channel] = FALSE;
	}

	return skipped;
}

gboolean sync_gtile_pixels_gimp(VGLayer *layer, GTile *gtile, GimpTile *tile, double curtime)
{
	int nsubtiles = gimp_tile_width()/VN_B_TILE_SIZE;
	int x, y, tx, ty, tw, th;
	int rwidth = (tile->ewidth/VN_B_TILE_SIZE)*VN_B_TILE_SIZE;
	int rheight = (tile->eheight/VN_B_TILE_SIZE)*VN_B_TILE_SIZE;
	VTile *vtile = gtile->tiles;
	gboolean skipped = FALSE;

	/* copy verse tiles into a gimp tile */
	ty = 0;
	th = VN_B_TILE_SIZE;
	for (y = 0; y < nsubtiles; y++, ty += VN_B_TILE_SIZE) {
		if (ty >= tile->eheight) break;
		else if (ty >= rheight) th = tile->eheight - rheight;

		tx = 0;
		tw = VN_B_TILE_SIZE;
		for (x = 0; x < nsubtiles; x++, vtile++, tx += VN_B_TILE_SIZE) {
			if (!(vtile->g.need_update[0] || vtile->g.need_update[1] ||
			      vtile->g.need_update[2] || vtile->g.need_update[3]))
				continue;
			else if (tx >= tile->ewidth) continue;
			else if (tx >= rwidth) tw = tile->ewidth - rwidth;

			skipped = skipped ||
			          sync_vtile_pixels_gimp(layer, vtile, tile, tx, ty, tw, th,
					                         curtime);
		}
	}

	return skipped;
}

void sync_vtile_pixels_verse(VGLayer *layer, VTile *vtile, GimpTile *tile, int tx, int ty, int tw, int th, int tilex, int tiley, double curtime)
{
	VGBitmap *bmp = layer->bitmap;
	int vtile_rowskip = VN_B_TILE_SIZE - tw;
	int tile_rowskip = (tile->ewidth - tw)*tile->bpp;
	int x, y, channel;
	uint8 *out;
	uint8 *in, *inptr = tile->data + (ty*tile->ewidth + tx)*tile->bpp;
	VNBTile vnbtile;

	/* verse expects all tile pixels to be zero outside the image */
	memset(vnbtile.vuint8, 0, sizeof vnbtile.vuint8);

	/* copy one verse tile into a gimp tile */
	for (channel = 0; channel < 4; channel++) {
		if (vtile->g.uninitialized[channel])
			continue;
		if (layer->v.layer_id[channel] == INVALID_LAYER_ID)
			continue;

		out = vnbtile.vuint8;
		in = inptr + channel;
		for (y = 0; y < th; y++) {
			for (x = 0; x < tw; x++, out++, in+=tile->bpp)
				*out = *in;

			out += vtile_rowskip;
			in += tile_rowskip;
		}

		if (memcmp(vnbtile.vuint8, vtile->g.data[channel], sizeof vnbtile.vuint8) == 0)
			continue;

		vtile->g.last_diff = curtime;
		memcpy(vtile->g.data[channel], vnbtile.vuint8, sizeof vnbtile.vuint8);
		verse_send_b_tile_set(bmp->v.node_id, layer->v.layer_id[channel],
		                      tilex, tiley, 0, VN_B_LAYER_UINT8, &vnbtile);
	}
}

void sync_gtile_pixels_verse(VGLayer *layer, GTile *gtile, GimpTile *tile, int xtile, int ytile, double curtime)
{
	int nsubtiles = gimp_tile_width()/VN_B_TILE_SIZE;
	int x, y, tx, ty, tw, th;
	int rwidth = (tile->ewidth/VN_B_TILE_SIZE)*VN_B_TILE_SIZE;
	int rheight = (tile->eheight/VN_B_TILE_SIZE)*VN_B_TILE_SIZE;
	VTile *vtile = gtile->tiles;

	/* copy verse tiles into a gimp tile */
	ty = 0;
	th = VN_B_TILE_SIZE;
	for (y = 0; y < nsubtiles; y++, ty += VN_B_TILE_SIZE) {
		if (ty >= tile->eheight) break;
		else if (ty >= rheight) th = tile->eheight - rheight;

		tx = 0;
		tw = VN_B_TILE_SIZE;
		for (x = 0; x < nsubtiles; x++, vtile++, tx += VN_B_TILE_SIZE) {
			if (tx >= tile->ewidth)
				continue;
			else if (tx >= rwidth)
				tw = tile->ewidth - rwidth;

			sync_vtile_pixels_verse(layer, vtile, tile, tx, ty, tw, th,
			                        xtile*nsubtiles + x, ytile*nsubtiles + y,
									curtime);
		}
	}
}

static void sync_layer_pixels(VGLayer *layer, uint32 difftiles)
{
	VGBitmap *bmp = layer->bitmap;
	GTile *gtile;
	int xtile, ytile, gsize = gimp_tile_width();
	GimpTile *tile;
	int xmin = bmp->width, ymin = bmp->height, xmax = 0, ymax = 0;
	GimpDrawable *drawable = gimp_drawable_get(layer->g.layer_id);
	gboolean skipped = FALSE, diff;
	double curtime = timer();

	/* copy verse layers into a gimp layer */
	for (ytile = 0; ytile < bmp->ytiles; ytile++) {
		for (xtile = 0; xtile < bmp->xtiles; xtile++) {
			gtile = &layer->tiles[xtile + ytile*bmp->xtiles];

			diff = (difftiles > 0);
			diff = diff && (layer->g.xdiff==xtile) && (layer->g.ydiff==ytile);
			if (!diff && !gtile->g.need_update)
				continue;

			tile = gimp_drawable_get_tile(drawable, FALSE, ytile, xtile);
			gimp_tile_ref(tile);

			/* WIP always diff before updating from verse! */
			if (gtile->g.need_update || diff) {
				sync_gtile_pixels_verse(layer, gtile, tile, xtile, ytile, curtime);

				if (difftiles != 0) {
					layer->g.xdiff = (layer->g.xdiff + 1) % bmp->xtiles;
					if (layer->g.xdiff == 0)
						layer->g.ydiff = (layer->g.ydiff + 1) % bmp->ytiles;
					difftiles--;
				}
			}
			if (gtile->g.need_update) {
				if (!sync_gtile_pixels_gimp(layer, gtile, tile, curtime))
					gtile->g.need_update = FALSE;
				else
					/* skipped is true if updating of a gimp image was delayed */
					skipped = TRUE;

				/* keep track of boundbox of updated tiles */
				if (xtile*gsize + tile->ewidth > xmax)
					xmax = xtile*gsize + tile->ewidth;
				if (ytile*gsize + tile->eheight > ymax)
					ymax = ytile*gsize + tile->eheight;
				if (xtile*gsize < xmin)
					xmin = xtile*gsize;
				if (ytile*gsize < ymin)
					ymin = ytile*gsize;

				gimp_tile_unref(tile, TRUE); /* tile dirty */
			}
			else
				gimp_tile_unref(tile, FALSE);
		}
	}

	gimp_drawable_update(layer->g.layer_id, xmin, ymin, xmax - xmin, ymax - ymin);
	gimp_drawable_detach(drawable);

	if (!skipped)
		layer->g.need_update = FALSE;
}

/* VGLayer */

static void sync_layer_alloc_tiles(VGLayer *layer, gboolean uploading)
{
	int nsubtiles = gimp_tile_width()/VN_B_TILE_SIZE;
	int i, j, ntiles = layer->bitmap->xtiles * layer->bitmap->ytiles;
	int channel;

	layer->tiles = calloc(ntiles, sizeof *(layer->tiles));

	for (i = 0; i < ntiles; i++) {
		GTile *gtile = &layer->tiles[i];
		gtile->tiles = calloc(nsubtiles*nsubtiles, sizeof *(gtile->tiles));

		if (!uploading)
			for (j = 0; j < nsubtiles*nsubtiles; j++)
				for(channel = 0; channel < 4; channel++)
					gtile->tiles[j].g.uninitialized[channel] = TRUE;

		/*for (j = 0; j < nsubtiles*nsubtiles; j++)
			memset(gtile->tiles[j].g.data, 255, sizeof gtile->tiles[j].g.data); */
	}
}

static void sync_layer_free_tiles(VGLayer *layer)
{
	int i, ntiles = layer->bitmap->xtiles * layer->bitmap->ytiles;

	if (!layer->tiles)
		return;

	for (i = 0; i < ntiles; i++)
		free(layer->tiles[i].tiles);
	free(layer->tiles);

	layer->tiles = NULL;
	layer->g.xdiff = layer->g.ydiff = 0;
}

static void sync_layer_free(VGLayer *layer)
{
	sync_layer_free_tiles(layer);

	if (layer->name)
		free(layer->name);

	free(layer);
}

VGLayer *sync_layer_lookup_verse(VGBitmap *bmp, VLayerID layer_id, int *channel)
{
	GList *i;
	VGLayer *layer;
	int j;
	
	for (i = bmp->layers; i; i = g_list_next(i)) {
		layer = (VGLayer*)i->data;

		for (j = 0; j < 4; j++) {
			if (layer->v.layer_id[j] == layer_id) {
				*channel = j;
				return layer;
			}
		}
	}

	*channel = -1;
	return NULL;
}

static VGLayer *sync_layer_create(VGBitmap *bmp, const char *name)
{
	VGLayer *layer = calloc(1, sizeof *layer);

	layer->v.layer_id[0] = layer->v.layer_id[1] = INVALID_LAYER_ID;
	layer->v.layer_id[2] = layer->v.layer_id[3] = INVALID_LAYER_ID;
	layer->v.nlayers = 0;
	layer->g.layer_id = -1;
	layer->name = strdup(name);

	layer->bitmap = bmp;
	bmp->layers = g_list_append(bmp->layers, layer);

	return layer;
}

static VGLayer *sync_layer_assign_verse(VGBitmap *bmp, VLayerID layer_id, const char *name, int *channel)
{
	GList *i;
	VGLayer *layer = NULL;

	/* assign a VGLayer to a verse layer if we are interested in it */

	*channel = - 1;
	
	if (strcmp(name, "color_r") == 0)
		*channel = 0;
	else if (strcmp(name, "color_g") == 0)
		*channel = 1;
	else if (strcmp(name, "color_b") == 0)
		*channel = 2;
	/*else if (strcmp(name, "alpha") == 0)
		*channel = 3;*/
	else
		return NULL;

	/* see if we already have a VGLayer for this verse layer / channel */
	for (i = bmp->layers; i; i = g_list_next(i)) {
		layer = (VGLayer*)i->data;

		if (layer->name && strcmp(layer->name, "main") == 0)
			break;
	}

	if (!i) /* otherwise create one */
		layer = sync_layer_create(bmp, "main");

	if (layer->v.layer_id[*channel] != INVALID_LAYER_ID) {
		fprintf(stderr, "duplicate channel");
		return NULL;
	}

	layer->v.layer_id[*channel] = layer_id;
	layer->v.nlayers++;

	return layer;
}

static gint32 sync_layer_find_main_gimp_layer(VGBitmap *bmp)
{
	gint32 *layers, layer_id;
	gint num_layers;
	gint i;

	layers = gimp_image_get_layers(bmp->g.image_id, &num_layers);

	for (i = 0; i < num_layers; i++)
		if (gimp_drawable_is_rgb(layers[i]))
			break;
	
	layer_id = (i == num_layers)? -1: layers[i];
	g_free(layers);

	return layer_id;
}

static void sync_layer_assign_gimp(VGLayer *layer, gboolean uploading)
{
	VGBitmap *bmp = layer->bitmap;
	gint32 layer_id;

	/* find a gimp layer for this VGLayer */

	if (layer->g.layer_id != -1)
		return;
	if (!layer->name || strcmp(layer->name, "main") != 0)
		return;

	layer_id = sync_layer_find_main_gimp_layer(bmp);

	if (layer_id != -1) {
		layer->g.sync = TRUE;
		layer->g.layer_id = layer_id;

		sync_layer_alloc_tiles(layer, uploading);
	}
}

static void sync_layer_unassign_gimp(VGLayer *layer)
{
	VGBitmap *bmp = layer->bitmap;
	int j;

	layer->g.sync = FALSE;
	layer->g.layer_id = -1;

	sync_layer_free_tiles(layer);

	for (j = 0; j < 4; j++) {
		if (layer->v.layer_id[j] == INVALID_LAYER_ID)
			continue;
		verse_send_b_layer_unsubscribe(bmp->v.node_id, layer->v.layer_id[j]);
	}
}

static gboolean sync_layer_verify(VGLayer *layer, gint32 *layers, gint num_layers)
{
	VGBitmap *bmp = layer->bitmap;
	gint i;

	/* verify if gimp layer is still in sync with VGLayer */

	if (layer->g.layer_id == -1)
		return TRUE;

	for (i = 0; i < num_layers; i++) {
		if (layer->g.layer_id == layers[i]) {
			if (gimp_drawable_width(layer->g.layer_id) != bmp->width ||
				gimp_drawable_height(layer->g.layer_id) != bmp->height)
				break;
			if (!gimp_drawable_is_rgb(layers[i]))
				break;

			/* WIP: no check for gimp_drawable_has_alpha(layers[i]) now */

			return TRUE;
		}
	}

	sync_layer_unassign_gimp(layer);

	return FALSE;
}

/* VGBitmap */

static void sync_bitmap_free(VGBitmap *bmp)
{
	GList *i;
	
	for (i = bmp->layers; i; i = g_list_next(i))
		sync_layer_free((VGLayer*)i->data);
	
	g_list_free(bmp->layers);

	if (bmp->name)
		free(bmp->name);

	free(bmp);
}

VGBitmap *sync_bitmap_lookup_verse(VNodeID node_id)
{
	GList *i;
	VGBitmap *bmp;
	
	for (i = G.sync.bitmaps; i; i = g_list_next(i)) {
		bmp = (VGBitmap*)i->data;

		if (bmp->v.sync && (bmp->v.node_id == node_id))
			return bmp;
	}

	return NULL;
}

VGBitmap *sync_bitmap_lookup_gimp(guint32 image_id)
{
	GList *i;
	VGBitmap *bmp;
	
	for (i = G.sync.bitmaps; i; i = g_list_next(i)) {
		bmp = (VGBitmap*)i->data;

		if (bmp->g.sync && (bmp->g.image_id == image_id))
			return bmp;
	}

	return NULL;
}

static gboolean sync_bitmap_exists(VGBitmap *bmp, gint32 *images, gint num_images)
{
	gint i;

	/* check if image still exists */
	for (i = 0; i < num_images; i++)
		if (bmp->g.image_id == images[i])
			return TRUE;
	
	return FALSE;
}

static gboolean sync_bitmap_layers_verify(VGBitmap *bmp)
{
	GList *i;
	gboolean valid = TRUE;
	gint num_layers;
	gint32 *layers = gimp_image_get_layers(bmp->g.image_id, &num_layers);

	for (i = bmp->layers; i; i = g_list_next(i))
		if (!sync_layer_verify((VGLayer*)i->data, layers, num_layers))
			valid = FALSE;
	
	g_free(layers);
	
	return valid;
}

/* Verse side syncing */

void sync_verse_bitmap_create(VNodeID node_id, VNodeOwner owner)
{
	GList *i;
	VGBitmap *bmp = sync_bitmap_lookup_verse(node_id);

	if (bmp)
		return;

	/* see if we have a gimp bitmap waiting for sync */
	if (owner == VN_OWNER_MINE) {
		for (i = G.sync.bitmaps; i; i = g_list_next(i)) {
			bmp = (VGBitmap*)i->data;

			if (bmp->g.sync && bmp->g.wait_sync) {
				bmp->g.wait_sync = FALSE;
				break;
			}
		}
	}
	
	/* otherwise create a new one */
	if (!bmp) {
		bmp = calloc(1, sizeof *bmp);
		G.sync.bitmaps = g_list_append(G.sync.bitmaps, bmp);
	}

	/* init verse properties */
	bmp->v.sync = TRUE;
	bmp->v.node_id = node_id;

	/* send gimp bitmap dimensions and name */
	if (bmp->g.sync) {
		gint32 drawable_id = sync_layer_find_main_gimp_layer(bmp);

		if (drawable_id != -1) {
			bmp->g.wait_dimensions = TRUE;

			bmp->width = gimp_drawable_width(drawable_id);
			bmp->height = gimp_drawable_height(drawable_id);

			verse_send_b_dimensions_set(bmp->v.node_id, bmp->width, bmp->height, 1);
			verse_send_node_name_set(bmp->v.node_id, bmp->name);
		}
	}

	verse_send_node_subscribe(node_id);

	gui_refresh();
}

void sync_verse_bitmap_dimensions(VGBitmap *bmp, int width, int height)
{
	if (bmp->g.sync) {
		if ((width == 0) && (height == 0))
			return;

		if (bmp->g.wait_dimensions) {
			gint32 drawable_id = sync_layer_find_main_gimp_layer(bmp);

			if (drawable_id != -1) {
				VGLayer *layer = sync_layer_create(bmp, "main");
				VNodeID node_id = bmp->v.node_id;

				bmp->g.wait_dimensions = FALSE;
				layer->g.wait_sync = 3;
				
				verse_send_b_layer_create(node_id, ~0, "color_r", VN_B_LAYER_UINT8);
				verse_send_b_layer_create(node_id, ~0, "color_g", VN_B_LAYER_UINT8);
				verse_send_b_layer_create(node_id, ~0, "color_b", VN_B_LAYER_UINT8);

				/*if (gimp_drawable_has_alpha(drawable_id)) {
					layer->g.wait_sync = 4;
					verse_send_b_layer_create(node_id, ~0, "alpha", VN_B_LAYER_UINT8);
				}*/
			}
		}
		else {
			/* WIP: stops syncing on unrequested size change now .. */
			sync_gimp_bitmap_unsubscribe(bmp);
		}
	}

	bmp->width = width;
	bmp->height = height;
	bmp->xtiles = ((width + gimp_tile_width() - 1) / gimp_tile_width());
	bmp->ytiles = ((height + gimp_tile_width() - 1) / gimp_tile_width());

	gui_refresh();
}

void sync_verse_bitmap_destroy(VGBitmap *bmp)
{
	sync_bitmap_free(bmp);
	G.sync.bitmaps = g_list_remove(G.sync.bitmaps, bmp);

	gui_refresh();
}

void sync_verse_bitmap_name_set(VGBitmap *bmp, const char *name)
{
	if (bmp->name)
		free(bmp->name);
	bmp->name = strdup(name);

	gui_refresh();
}

void sync_verse_layer_create(VGBitmap *bmp, VLayerID layer_id, const char *name)
{
	int channel;
	VGLayer *layer = sync_layer_assign_verse(bmp, layer_id, name, &channel);

	if (layer) {
		if (layer->g.wait_sync) {
			/* we are uploading this bitmap */
			sync_layer_assign_gimp(layer, TRUE);
			verse_send_b_layer_subscribe(bmp->v.node_id, layer_id, 0);
			layer->g.wait_sync--;
		}
		else if (bmp->g.sync) {
			/* subscribing to this bitmap */
			sync_layer_assign_gimp(layer, FALSE);
			verse_send_b_layer_subscribe(bmp->v.node_id, layer_id, 0);
		}
	}
}

void sync_verse_layer_destroy(VGBitmap *bmp, VGLayer *layer, int channel)
{
	layer->v.layer_id[channel] = INVALID_LAYER_ID;
	layer->v.nlayers--;

	if (layer->v.nlayers == 0) {
		sync_layer_free(layer);
		bmp->layers = g_list_remove(bmp->layers, layer);
	}
}

void sync_verse_layer_tile_set(VGLayer *layer, int channel, int x, int y, uint8 *data)
{
	VGBitmap *bmp = layer->bitmap;
	int subx, suby, nsubtiles = gimp_tile_width()/VN_B_TILE_SIZE;
	GTile *gtile;
	VTile *vtile;

	if (!layer->tiles)
		return;

	subx = x / nsubtiles;
	suby = y / nsubtiles;

	if (subx >= bmp->xtiles || suby >= bmp->ytiles) {
		fprintf(stderr, "rejected tile\n");
		return;
	}
	
	gtile = &layer->tiles[subx + suby*bmp->xtiles];

	x = x - subx*nsubtiles;
	y = y - suby*nsubtiles;

	vtile = &gtile->tiles[x + y*nsubtiles];

	/* uploads take priority over downloads */
	vtile->g.need_update[channel] = TRUE;
	gtile->g.need_update = TRUE;
	layer->g.need_update = TRUE;

	memcpy(vtile->v.data[channel], data, sizeof vtile->v.data[channel]);
}

/* Gimp side syncing */

static void sync_gimp_bitmap_create(gint32 image_id)
{
	VGBitmap *bmp = calloc(1, sizeof *bmp);
	char *name = gimp_image_get_name(image_id);

	/* init verse properties */
	bmp->g.sync = TRUE;
	bmp->g.image_id = image_id;

	if (name) {
		bmp->name = strdup(name);
		g_free(name);
	}

	/* WIP: layer_id, width, height, .. ? check for upload syncing */

	G.sync.bitmaps = g_list_append(G.sync.bitmaps, bmp);

	gui_refresh();
}

void sync_gimp_bitmaps()
{
	GList *i;
	gint32 *images;
	gint num_images, j;
	gboolean removed;
	
	images = gimp_image_list(&num_images);

	/* remove or unsubscribe removed gimp images */
	do {
		removed = FALSE;

		for (i = G.sync.bitmaps; i; i = g_list_next(i)) {
			VGBitmap *bmp = (VGBitmap*)i->data;

			if (bmp->g.sync && !sync_bitmap_exists(bmp, images, num_images)) {
				if (!bmp->v.sync) {
					sync_bitmap_free(bmp);
					G.sync.bitmaps = g_list_remove(G.sync.bitmaps, bmp);
					removed = TRUE;
					gui_refresh();
					break;
				}
				else
					sync_gimp_bitmap_unsubscribe(bmp);
			}
		}
	} while (removed);

	/* unsubscibe gimp images without proper layers */
	for (i = G.sync.bitmaps; i; i = g_list_next(i)) {
		VGBitmap *bmp = (VGBitmap*)i->data;

		if (bmp->g.sync && bmp->v.sync)
			if (!sync_bitmap_layers_verify(bmp))
				sync_gimp_bitmap_unsubscribe(bmp);
	}

	/* add new gimp images */
	for (j = 0; j < num_images; j++)
		if (!sync_bitmap_lookup_gimp(images[j]))
			sync_gimp_bitmap_create(images[j]);

	g_free(images);
}

void sync_pixels(uint32 difftiles)
{
	GList *i, *j;
	VGLayer *layer;
	VGBitmap *bmp;
	gboolean updated = FALSE;

	/* write verse tiles to gimp image */

	for (i = G.sync.bitmaps; i; i = g_list_next(i)) {
		bmp = (VGBitmap*)i->data;

		if (!bmp->g.sync)
			continue;

		for (j = bmp->layers; j; j = g_list_next(j)) {
			layer = (VGLayer*)j->data;

			if (!layer->g.sync || !layer->tiles ||
			    (!layer->g.need_update && difftiles==0))
				continue;

			updated = TRUE;
			sync_layer_pixels(layer, difftiles);
		}
	}

	if (updated)
		gimp_displays_flush();
}

void sync_gimp_bitmap_sync(VGBitmap *bmp)
{
	if (bmp->g.wait_sync == FALSE) {
		bmp->g.wait_sync = TRUE;
		verse_send_node_create(~0, V_NT_BITMAP, 0);
	}
}

void sync_gimp_bitmap_destroy(VGBitmap *bmp)
{
	verse_send_node_destroy(bmp->v.node_id);
}

void sync_gimp_bitmap_subscribe(VGBitmap *bmp)
{
	GList *i;
	VGLayer *layer;
	gint32 layer_id, j;
	
	if ((bmp->g.sync) || (bmp->width == 0) || (bmp->height == 0))
		return;

	bmp->g.sync = TRUE;
	bmp->g.image_id = gimp_image_new(bmp->width, bmp->height, GIMP_RGB);
	gimp_image_set_filename(bmp->g.image_id, bmp->name);
	layer_id = gimp_layer_new(bmp->g.image_id, "Background", bmp->width,
	                          bmp->height, GIMP_RGB_IMAGE, 100, GIMP_NORMAL_MODE);
	gimp_image_add_layer(bmp->g.image_id, layer_id, 0);
#if 0
	/* WIP should check for alpha channel here and set to 0 or 255 according to
	   if an alpha layer exists */
#endif
	gimp_drawable_fill(layer_id, GIMP_WHITE_FILL);
	gimp_display_new(bmp->g.image_id);

	/* find gimp layers and subscribe to verse layers */
	for (i = bmp->layers; i; i = g_list_next(i)) {
		layer = (VGLayer*)i->data;

		sync_layer_assign_gimp(layer, FALSE);

		for (j = 0; j < 4; j++)
			if (layer->v.layer_id[j] != INVALID_LAYER_ID)
				verse_send_b_layer_subscribe(bmp->v.node_id, layer->v.layer_id[j], 0);
	}

	gui_refresh();
}

void sync_gimp_bitmap_unsubscribe(VGBitmap *bmp)
{
	GList *i;
	VGLayer *layer;
	
	for (i = bmp->layers; i; i = g_list_next(i)) {
		layer = (VGLayer*)i->data;
		sync_layer_unassign_gimp(layer);
	}

	bmp->g.sync = FALSE;

	gui_refresh();
}

/* Init */

void sync_start()
{
	G.sync.bitmaps = NULL;
}

void sync_stop()
{
	GList *i;
	VGBitmap *bmp;
	
	for (i = G.sync.bitmaps; i; i = g_list_next(i)) {
		bmp = (VGBitmap*)i->data;
		sync_bitmap_free(bmp);
	}

	g_list_free(G.sync.bitmaps);
	G.sync.bitmaps = NULL;
}

