
#include <verse.h>
#include <gtk/gtk.h>
#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

#define INVALID_LAYER_ID ((VLayerID)(-1))

/* Structs */

typedef struct VGBitmap {
	/* Represents a bitmap, that exists on the verse server, in the gimp,
	   or both, as indicated by v.sync and g.sync. */

	struct VBitmap {
		gboolean sync;
		VNodeID node_id;
	} v;

	struct GBitmap {
		gboolean sync;
		gboolean wait_sync;
		gboolean wait_dimensions;
		gint32 image_id;
	} g;

	gchar *name;
	int width, height;
	int xtiles, ytiles;
	GList *layers;
} VGBitmap;

typedef struct VTile {
	struct VTileV {
		/* incoming data to be written to the gimp */
		uint8 data[4][VN_B_TILE_SIZE*VN_B_TILE_SIZE];
		/* data in VTileG needs to be sent to verse, WIP, this is always
		   false now */
		gboolean need_update[4];
	} v;
	struct VTileG {
		/* latest gimp data for diffing tiles */
		uint8 data[4][VN_B_TILE_SIZE*VN_B_TILE_SIZE];
		/* gimp tile needs to be updated with data in VTileV */
		gboolean need_update[4];
		/* last time a difference on a gimp tile was detected */
		double last_diff;
		gboolean uninitialized[4];
	} g;

	/* this contains data received from verse */
} VTile;

typedef struct GTile {
	struct GTileV {
		gboolean need_update;
	} v;
	struct GTileG {
		gboolean need_update;
	} g;

	VTile *tiles;
} GTile;

typedef struct VGLayer {
	/* Represents a gimp layer with 1-4 verse layers / gimp channels */

	struct VLayer {
		VLayerID layer_id[4];
		int nlayers;
		gboolean need_update;
	} v;

	struct GLayer {
		gboolean sync;
		int wait_sync;
		gboolean need_update;
		gint32 layer_id;
		int xdiff, ydiff;
	} g;

	VGBitmap *bitmap;
	char *name;
	GTile *tiles;
} VGLayer;

enum ConnectionState {
	UNINITIALIZED,
	CONNECTING,
	CONNECTED,
	DISCONNECTED
};

struct Global {
	struct Gui {
		GtkWidget *window, *notebook;
		
		GtkWidget *ms_address;
		GtkListStore *ms_store;
		GtkWidget *ms_list_view;
		GtkWidget *ms_list;

		GtkWidget *table_server, *state_label;
		GtkWidget *pass, *uname, *address;
		GtkWidget *conn, *disconn;

		GtkListStore *gimplist, *verselist;
		GtkWidget *gimptree, *versetree;

		enum ConnectionState state;
		gboolean need_refresh;
	} gui;

	struct Connection {
		VSession *session;
	} conn;
	
	struct Sync {
		GList *bitmaps;
	} sync;
};

extern struct Global G;

/* GUI */

void gui_run();
void gui_set_conn_state(enum ConnectionState state);
enum ConnectionState gui_get_conn_state();
void gui_refresh();

/* Verse */

void conn_start(const char *addr, const char *uname, const char *pass);
void conn_stop();
void conn_callback_update();

/* Sync */

void sync_start();
void sync_stop();

void sync_verse_bitmap_create(VNodeID node_id, VNodeOwner owner);
void sync_verse_bitmap_dimensions(VGBitmap *bmp, int width, int height);
void sync_verse_bitmap_destroy(VGBitmap *bmp);
void sync_verse_bitmap_name_set(VGBitmap *bmp, const char *name);
void sync_verse_layer_create(VGBitmap *bmp, VLayerID layer_id, const char *name);
void sync_verse_layer_destroy(VGBitmap *bmp, VGLayer *layer, int channel);
void sync_verse_layer_tile_set(VGLayer *layer, int channel, int x, int y, uint8 *data);

void sync_gimp_bitmaps();
void sync_gimp_bitmap_sync(VGBitmap *bmp);
void sync_gimp_bitmap_destroy(VGBitmap *bmp);
void sync_gimp_bitmap_subscribe(VGBitmap *bmp);
void sync_gimp_bitmap_unsubscribe(VGBitmap *bmp);

VGBitmap *sync_bitmap_lookup_gimp(guint32 image_id);
VGBitmap *sync_bitmap_lookup_verse(VNodeID node_id);
VGLayer *sync_layer_lookup_verse(VGBitmap *bmp, VLayerID layer_id, int *channel);

void sync_pixels(uint32 difftiles);

